import os
import copy
import appdirs

from . import io

try:
  import yaml
except:
  pass

_DEFAULT_CFG = dict(
  debug = False,
  max_directory_depth = 5,
  graphics_extensions = ['jpg', 'jpeg', 'pdf', 'png'],
  img_dir_name = 'img',
  img_search_paths = [],
  img_search_priority = 'path-order, newest',
  input_search_paths = [],
  input_search_priority = 'path-order, newest',
  bibtex_extensions = ['bib', 'bibtex'],
  bib_search_paths = [],
  bib_search_priority = 'path-order, newest',
  bib_repair = dict(
    make_all_items_lowercase = True,
    merge_all_found_info = True,
    remove_fields = ['abstract'],
    generate_url_from_doi = True,
    force_double_hyphen_in_pages = True,
    keep_only_startpage_in_pages = False,
    convert_month_to_number = True,
    convert_journal_to_iso4_abbr = True,
    convert_journal_to_full_name = False,
    protect_capitalization_if_unprotected = True
  ),
  lint=dict(
    avoid_commands_in_toplevel=[],
    avoid_commands = [],
#    check_spelling=[]
  ),
  library_path = '',
  library_collect_paths = [],
  library_folder_pattern = '%Y-%m',
  library_max_filename_len = 60,
  library_max_filename_segment_len = 60,
  library_max_key_len = 20,
  library_max_key_segment_len = 5,
  z_bib_words_protect_capitalization = [],
  z_bib_words_dont_protect_capitalization = [],
  z_bib_journals = {}
)

_REQUIRED_CFG = dict()

_CFG_PATH = os.path.join(appdirs.user_config_dir('paperman'), 'paperman.conf')
_CFG = {}
_IS_CFG_LOADED = False
_HAS_TESTED_REQUIRED_KEYS = False


def _touchCfg():
  os.makedirs(os.path.dirname(_CFG_PATH), exist_ok=True)
  if not os.path.exists(_CFG_PATH):
    open(_CFG_PATH, 'w')


def _loadCfg():
  global _IS_CFG_LOADED, _CFG, _HAS_TESTED_REQUIRED_KEYS
  if not _IS_CFG_LOADED:
    _touchCfg()
    _CFG = yaml.safe_load(open(_CFG_PATH, 'r')) or {}
    _IS_CFG_LOADED = True
    _HAS_TESTED_REQUIRED_KEYS = False
    testIfRequiredExist()
    _writeCfg()


def _writeCfg():
  _touchCfg()
  yaml.dump({k: v for k, v in {**_DEFAULT_CFG, **_CFG}.items()},
            open(_CFG_PATH, 'w'))


def reload():
  global _IS_CFG_LOADED
  try:
    _IS_CFG_LOADED = False
    _loadCfg()
  except (KeyboardInterrupt, RuntimeError):
    raise
  except:
    io.err(io.coercedStacktrace(10))
    io.err('error reloading config file, reverting '
           'changes on disk')
    _writeCfg()
    _loadCfg()


def testIfRequiredExist(*path):
  global _HAS_TESTED_REQUIRED_KEYS, _IS_CFG_LOADED
  if not _HAS_TESTED_REQUIRED_KEYS:
    settings = _REQUIRED_CFG
    for p in path:
      settings = settings[p]
    if type(settings) is dict:
      for k in settings.keys():
        testIfRequiredExist(*path, k)
    else:
      if get(*path) is None:
        _IS_CFG_LOADED = False
        raise RuntimeError(f'config setting {".".join(path)} is required '
                           f'but is not set')
  if len(path) == 0:
    _HAS_TESTED_REQUIRED_KEYS = True


def get(*key):
  if len(key) < 1:
    raise ArgumentError('cfg.get() expects at least one argument: key')
  _loadCfg()
  setting, default = _CFG, _DEFAULT_CFG
  usedDefault = False
  for k in key:
    if k not in setting.keys():
      usedDefault = True
      setting[k] = default[k]
    default = default[k]
    setting = setting[k]

  if usedDefault:
    _writeCfg()

  return setting


def set(*keyVal):
  if len(keyVal) < 2:
    raise ArgumentError('cfg.set() expects at least two arguments: key and value')
  key = keyVal[:-1]
  val = nativify(keyVal[-1])
  reload()
  d = _CFG
  for k in key[:-1]:
    if k not in d:
      d[k] = {}
    d = d[k]
  d[key[-1]] = val
  _writeCfg()


def nativify(data, skipIfFunc=lambda data: False, skipIfError=False):
  args = dict(skipIfFunc=skipIfFunc, skipIfError=skipIfError)

  if data is None or type(data) in (bool, int, float, complex, str):
    return data

  if type(data) is dict:
    return {nativify(k, **args): nativify(v, **args) for k, v in data.items()}

  if hasattr(data, '__iter__'):
    try:
      return [nativify(v, **args) for v in data]
    except:
      pass

  try:
    return float(data)
  except:
    try:
      return complex(data)
    except:
      pass

  if skipIfError or skipIfFunc(data):
    return None

  raise ValueError('could not convert data to native type: "'+str(data)+'"')



import re
import time
import shutil

from .. import parser
from . common import *


def makeKey(string, maxTotLen, maxSegLen):
  res = ""
  for s in re.sub("[^a-z0-9 ]+", "", string.lower()).split():
    s = s.lower().strip()
    if len(res)+len(s[:maxSegLen])+1 > maxTotLen:
      break
    elif res == "":
      res += s[:maxSegLen]
    else:
      res += "-"+s[:maxSegLen]
  return res


def main(args):
  # check of library path is set
  libraryPath = os.path.expanduser(cfg.get('library_path'))
  if not libraryPath:
    io.err('library_path is not set in config file')
    return

  # check if collect paths exist
  collectPaths = args.paths or cfg.get('library_collect_paths')
  if collectPaths:
    io.verb('colling bib-pdf pairs from the follwoing paths:', *collectPaths)
  else:
    io.err('no collect paths specified, either',
           'set "library_collect_paths" in the config or',
           'pass paths as argument to "paperman collect"')
    return

  # define routine to use for error reporting
  def onError(file, *err):
    if args.err_to_file:
      errStr = io.formatErr(f'failed to collect "{file}":', *err)
      with open(os.path.join(os.path.dirname(file),
                             'paperman-collect-error.txt'), 'w') as f:
        f.write(errStr)
    io.err(*err)

  # mainloop
  while True:
    try:
      for p in collectPaths:
        p = os.path.expanduser(p)

        pdfs = [os.path.join(p, f) for f in os.listdir(p)
                          if os.path.isfile(os.path.join(p, f))
                              and f.lower().endswith('.pdf')]

        bibs = [os.path.join(p, f) for f in os.listdir(p)
                          if os.path.isfile(os.path.join(p, f))
                              and any([f.lower().endswith(e)
                                  for e in list(cfg.get('bibtex_extensions'))
                                                  +['.ris']])]

        io.verb('in folder ', p,
                'found pdf candidates: ', *(pdfs or ['-none-']),
                'found bib candidates: ', *(bibs or ['-none-']))

        if len(pdfs) == 1 and len(bibs) == 1:
          pdfPath = pdfs[0]
          bibPath = bibs[0]

          # load and parse bib file
          if bibPath.lower().endswith('.ris'):
            cites = parser.BibFile.fromRis('key', bibPath).cites()
          else:
            cites = parser.BibFile(bibPath).cites()

          if len(cites) == 1:
            cite = cites[0]

            fullTitle = cite['title']
            if fullTitle:

              # generate key and filename from paper title
              libraryName = makeKey(fullTitle,
                                    cfg.get('library_max_filename_len'),
                                    cfg.get('library_max_filename_segment_len'))
              cite.key = makeKey(fullTitle,
                                 cfg.get('library_max_key_len'),
                                 cfg.get('library_max_key_segment_len'))
              libraryDir = os.path.join(libraryPath,
                                        time.strftime(cfg.get('library_folder_pattern')))
              target = os.path.join(libraryDir, libraryName)
              os.makedirs(os.path.dirname(target), exist_ok=True)
              if not any([f.startswith(target)
                              for f in os.listdir(os.path.dirname(target))]):
                shutil.move(pdfPath, target+'.pdf')
                io.info(f'moved "{pdfPath}" -> "{target}.pdf"')
                with open(target+'.'+cfg.get('bibtex_extensions')[0], 'w') as f:
                  f.write(cite.pretty()+'\n')
                os.remove(bibPath)
                io.info(f'moved "{bibPath}" -> "{target}.{cfg.get("bibtex_extensions")[0]}"')
              else:
                onError(pdfPath, 'failed to move files to library, target',
                        '"'+target+'"', 'already exists')

            else:
              onError(pdfPath, 'failed to find title in bib file')

          else:
            onError(pdfPath, f'found unexpected citation count ({len(cites)}) in bib file')

        elif len(pdfs) or len(bibs):
          onError(os.path.join(p, 'derp'),
                  f'found {len(pdfs)} pdf candidates and {len(bibs)}',
                  f'bib candidates in ', p,
                  f'need to have exactly one of each to collect')

      # if watch option is not set run mainloop only once
      if not args.watch:
        break
      time.sleep(5)

    except KeyboardInterrupt:
      raise
    except:
      if not args.watch:
        raise
      io.err('unexpected error:', io.coercedStacktrace(30))
      time.sleep(5)
